import hashlib
import os
from tqdm import tqdm
from google.cloud import storage
import click


class googleStorage:
    download_dir = "download-model-ver"

    def __init__(self, bucket_name):
        self.storage_client = storage.Client()
        self.bucket = self.storage_client.get_bucket(bucket_name)
        self.bucket_name = bucket_name

    def delete_entire_directory(self, directory_name):
        blobs = self.bucket.list_blobs(prefix=directory_name)
        for blob in blobs:
            blob.delete()
        click.echo("ver deleted successfully")

    def delete_file(self, directory_name, file_name):
        hash_file = (file_name.split('.'))[0] + '.hash'
        blob = self.bucket.blob(directory_name + file_name)
        blob.delete()
        blob = self.bucket.blob(directory_name + hash_file)
        blob.delete()
        click.echo("file deleted successfully")

    def download_file_from_storage(self, source_blob_name, destination_file_name):
        destination_full_file_name = os.getcwd() + '/' + self.download_dir + '/' + destination_file_name
        if self.file_exists(source_blob_name):
            dirname = os.path.dirname(destination_full_file_name)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            blob = self.bucket.get_blob(source_blob_name)
            storage.blob._MAX_MULTIPART_SIZE = 5 * 1024 * 1024
            blob._chunk_size = 5 * 1024 * 1024
            with open(destination_full_file_name, "wb") as in_file:
                total_bytes = blob.size
                with tqdm.wrapattr(in_file, "write", total=total_bytes, miniters=1,
                                   desc="download to %s" % destination_full_file_name, colour="green") as file_obj:
                    self.storage_client.download_blob_to_file(blob, file_obj)
        else:
            click.echo(
                "object {} from bucket {} not found".format(
                    source_blob_name, self.bucket_name
                )
            )

    def download_ver_from_storage(self, project, ver):
        blob_list = self.list_dir(project + "/" + ver, True, False)
        for blob in blob_list:
            blobItemsArray = blob.split('/')
            i = 0
            destFileName = ""
            for blobItem in blobItemsArray:
                if (i > 1):
                    destFileName = destFileName + blobItem + '/'
                i = i + 1
            destFileName = destFileName[:-1]
            self.download_file_from_storage(blob, ver + '/' + destFileName)

    def download_all_ver_from_storage(self, project):
        ver_list = self.list_dir(project + "/", False, False)
        for ver in ver_list:
            ver = ver[:-1]
            self.download_ver_from_storage(project, ver)

    def list_dir(self, full_folder, get_version_files, display_console=True):
        if get_version_files:
            blobs = self.storage_client.list_blobs(self.bucket_name, prefix=full_folder)
            blob_list = []
            for blob in blobs:
                if not blob.name.endswith('.hash'):
                    blob_list.append(blob.name)
                    blob_no_prefix = blob.name.replace(full_folder, "")
                    if display_console:
                        click.echo(blob_no_prefix)
            return blob_list
        else:
            blobs = self.storage_client.list_blobs(
                bucket_or_name=self.bucket_name,
                prefix=full_folder,
                delimiter="/",
                max_results=1
            )
            next(blobs, ...)
            blob_list = []
            for folder in list(blobs.prefixes):
                dir_no_suffix = folder.replace(full_folder, "")
                blob_list.append(dir_no_suffix)
                if display_console:
                    click.echo(dir_no_suffix)
            return blob_list

    def upload_to_bucket(self, full_file_path, path_to_file, bucket_hash_key=None):
        if not self.file_exists(full_file_path):
            self.upload_file_and_hash_to_bucket(full_file_path, path_to_file)
        else:
            bucket_hash_file = self.rename_file_to_hash_suffix(full_file_path)
            if self.file_exists(bucket_hash_file):
                bucket_hash_key = self.get_file(bucket_hash_file).decode('utf-8')
            calculated_hash = self.calculate_hash(path_to_file)
            if calculated_hash != bucket_hash_key:
                overwrite_question = input(
                    "a different file with the same name already exists do you want to overwrite it (yes, no)\n")
                if overwrite_question.lower() == ("yes"):
                    self.upload_file_and_hash_to_bucket(full_file_path, path_to_file)

    def upload_file_and_hash_to_bucket(self, full_file_path, path_to_file, content_type=None):
        blob = self.bucket.blob(full_file_path)
        storage.blob._MAX_MULTIPART_SIZE = 5 * 1024 * 1024
        blob._chunk_size = 5 * 1024 * 1024
        with open(path_to_file, "rb") as in_file:
            total_bytes = os.fstat(in_file.fileno()).st_size
            with tqdm.wrapattr(in_file, "read", total=total_bytes, miniters=1,
                               desc="upload to %s" % full_file_path, colour="green") as file_obj:
                blob.upload_from_file(
                    file_obj,
                    content_type=content_type,
                    size=total_bytes,
                )
                self.create_hash_and_upload(full_file_path, path_to_file)

    def create_hash_and_upload(self, full_file_path, path_to_file):
        hash_file = self.rename_file_to_hash_suffix(full_file_path)
        calculated_hash = self.calculate_hash(path_to_file)
        blob = self.bucket.blob(hash_file)
        blob.upload_from_string(calculated_hash)

    def rename_file_to_hash_suffix(self, file_name):
        return os.path.splitext(file_name)[0] + '.hash'

    def get_file(self, full_file_path):
        blob = self.bucket.blob(full_file_path)
        data = blob.download_as_string()
        return data;

    # def get_files(self, full_folder):
    #     blobs = self.bucket.list_blobs(prefix=full_folder)
    #     for blob in blobs:
    #         print(blob.name)

    def calculate_hash(self, file_path):
        BUF_SIZE = 65536
        md5 = hashlib.md5()
        with open(file_path, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                md5.update(data)
        return md5.hexdigest()

    def file_exists(self, full_file_path):
        return storage.Blob(bucket=self.bucket, name=full_file_path).exists(self.storage_client)



import json
import os
import click
from os import walk

from .google_storage import googleStorage


class App:
    __conf = {
        "Chosen_Cloud_Provider": "", "Google_Bucket_Name": "",
        "AWS_Bucket_Name": ""
    }
    __setters = ["Chosen_Cloud_Provider", "Google_Bucket_Name", "AWS_Bucket_Name"]

    @staticmethod
    def config(name):
        return App.__conf[name]

    @staticmethod
    def set(name, value):
        if name in App.__setters:
            App.__conf[name] = value
        else:
            raise NameError("Name not accepted in set() method")


class func:
    config_dir = "config-model-ver"
    full_path_config_file = config_dir + '/config.json'

    def __init__(self, init=True):
        if init:
            self.config_files_exists = func.check_config_files_exists(self)
            self.get_bucket_name()
            self.func_storage = googleStorage(self.bucket_name)

    def delete_entire_directory(self, directory_name):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.delete_entire_directory(directory_name)

    def delete_file(self, directory_name, file_name):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.delete_file(directory_name, file_name)

    def download_file_from_storage(self, source_blob_name, destination_file_name):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.download_file_from_storage(source_blob_name, destination_file_name)

    def download_ver_from_storage(self, project, ver):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.download_ver_from_storage(project, ver)

    def download_all_ver_from_storage(self, project):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.download_all_ver_from_storage(project)

    def list_dir(self, full_folder, get_version_files, display_console=True):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        self.func_storage.list_dir(full_folder, get_version_files, display_console)

    def add_directory(self, directory, project, ver):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        full_directory = os.getcwd() + '/' + directory
        filenames = next(walk(full_directory), (None, None, []))[2]
        for file in filenames:
            self.upload_file(project, ver, full_directory + '/' + file, directory)

    def upload_file(self, project, ver_name, file_name, dir_name):
        storage_file_name = os.path.basename(file_name)
        if os.path.exists(file_name):
            if dir_name == '':
                self.func_storage.upload_to_bucket(project + '/' + ver_name + '/' + storage_file_name, file_name)
            else:
                self.func_storage.upload_to_bucket(project + '/' + ver_name + '/' + dir_name + '/' + storage_file_name,
                                                   file_name)
        else:
            click.echo("file " + file_name + " does not exists")

    def add_file(self, project, ver_name, file_name):
        if self.bucket_name == "" or not self.config_files_exists:
            return
        if not file_name.endswith("/."):
            self.upload_file(project, ver_name, file_name, '')
        else:
            prefix_dir = file_name.replace("/.", "")
            if len(prefix_dir) > 0:
                prefix_dir = prefix_dir + "/"
            full_dir_name = os.getcwd() + '/' + prefix_dir
            filenames = next(walk(full_dir_name), (None, None, []))[2]
            for file in filenames:
                self.upload_file(project, ver_name, prefix_dir + file, '')

    def get_bucket_name(self):
        if App.config("Chosen_Cloud_Provider") == "google":
            self.bucket_name = App.config("Google_Bucket_Name")
        elif App.config("Chosen_Cloud_Provider") == "aws":
            self.bucket_name = App.config("Google_Bucket_Name")
        if self.bucket_name == "":
            click.echo("no bucket defined, please configure bucket name")

    def create_config_if_not_exists(self):
        full_path_config_file = self.config_dir + '/config.json'
        if not os.path.exists(self.config_dir):
            os.makedirs(self.config_dir)

        if not os.path.isfile(full_path_config_file):
            json_create_first_time = {"Chosen_Cloud_Provider": "", "Google_Bucket_Name": "",
                                      "AWS_Bucket_Name": ""}
            with open(full_path_config_file, 'w') as config_file:
                json.dump(json_create_first_time, config_file)
            return False
        else:
            return True

    def save_cloud_provider_to_config_file(self, cloud_provider):
        self.create_config_if_not_exists()
        with open(self.full_path_config_file) as config_file:
            json_decoded = json.load(config_file)
        json_decoded['Chosen_Cloud_Provider'] = cloud_provider
        with open(self.full_path_config_file, 'w') as config_file:
            json.dump(json_decoded, config_file)
            click.echo("cloud provider set successfully")

    def save_config_file(self, cloud_provider, bucket_name):
        config_file_exists = self.create_config_if_not_exists()
        with open(self.full_path_config_file) as config_file:
            json_decoded = json.load(config_file)
        if not config_file_exists:
            json_decoded['Chosen_Cloud_Provider'] = cloud_provider
        if cloud_provider.lower() == "google":
            json_decoded['Google_Bucket_Name'] = bucket_name
        elif cloud_provider.lower() == "aws":
            json_decoded['AWS_Bucket_Name'] = bucket_name
        with open(self.full_path_config_file, 'w') as config_file:
            json.dump(json_decoded, config_file)
            click.echo("configuration file saved successfully")

    def display_config_file(self):
        if not self.read_config_file(True):
            click.echo("no configuration file was found, please use configure flag to create one")

    def check_config_files_exists(self):
        if not self.read_config_file():
            return False
        else:
            return True

    def read_config_file(self, print_json=False):
        if os.path.exists(self.config_dir + '/config.json'):
            with open('config-model-ver/config.json', 'r') as config_file:
                json_config = json.load(config_file)
                if "Chosen_Cloud_Provider" in json_config:
                    App.set("Chosen_Cloud_Provider", json_config["Chosen_Cloud_Provider"])
                if "Google_Bucket_Name" in json_config:
                    App.set("Google_Bucket_Name", json_config["Google_Bucket_Name"])
                if "AWS_Bucket_Name" in json_config:
                    App.set("AWS_Bucket_Name", json_config["AWS_Bucket_Name"])
                if print_json:
                    pretty_json = json.dumps(json_config, indent=4)
                    click.echo(pretty_json)
            return True;
        else:
            print("Configuration file not found in config-model-ver directory\n")
            return False



import click

from py_model_versioning.classes.func import func


@click.group()
def cli():
    pass


@cli.command('print_ver', help="display ver for project")
def print_ver():
    click.echo("0.5")


@cli.command('del_ver', help="delete ver for project")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name")
def del_ver(p, v):
    call = func()
    call.delete_entire_directory(p + "/" + v + "/")


@cli.command('del_file', help="delete specific file from storage")
@click.option('-f', required=True, help="specify file name to delete")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name")
def del_file(f, p, v):
    call = func()
    call.delete_file(p + "/" + v + "/", f)


@cli.command('get_file', help="get specific fie from storage")
@click.option('-f', required=True, help="specify file name to get")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name")
def get_file(f, p, v):
    call = func()
    call.download_file_from_storage(p + "/" + v + "/" + f, f)


@cli.command('get_ver', help="get complete version files for specific project")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name or all for all versions")
def get_ver(p, v):
    call = func()
    if v != "all":
        call.download_ver_from_storage(p, v)
    else:
        call.download_all_ver_from_storage(p)


@cli.command('init', help="configure cloud provider bucket name and project for google, should be done from root directory")
@click.option('-c', required=True, help="specify cloud provider google or aws currently support only google")
@click.option('-b', required=True, help="specify bucket name")
def init(c, b):
    call = func(False)
    if c.lower() != "google" and c.lower() != "aws":
        click.echo("cloud provider can be google or aws")
    else:
        call.save_config_file(c, b)

@cli.command('init_cloud', help="config")
@click.option('-c', required=True, help="specify cloud provider google or aws currently support only google")
def init_cloud(c):
    call = func(False)
    if c.lower() != "google" and c.lower() != "aws":
        click.echo("cloud provider can be google or aws")
    else:
        call.save_cloud_provider_to_config_file(c)


@cli.command('display_init', help="display the content of configuration file")
def display_init():
    call = func(False)
    call.display_config_file()


@cli.command('ls', help="list storage content, no argument list projects")
@click.option('-p', required=False, help="list all project versions")
@click.option('-v', required=False, help="accomapny with -p list version content")
def ls(p, v):
    call = func()
    if p == None and v == None:
        call.list_dir('', False)
    elif p != None and v == None:
        call.list_dir(p + "/", False)
    else:
        call.list_dir(p + "/" + v, True)


@cli.command('add_directory', help="add directory files to project model version storage")
@click.option('-d', required=True, help="specify directory name to add")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name")
def add_directory(d, p, v):
    call = func()
    call.add_directory(d, p, v)


@cli.command('add_file', help="add files to project model version storage")
@click.option('-f', required=True, help="specify file name to add or /. to add all files within the folder")
@click.option('-p', required=True, help="specify project name")
@click.option('-v', required=True, help="specify version name")
def add_file(f, p, v):
    call = func()
    call.add_file(p, v, f)


if __name__ == "__main__":
    cli()
